//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                         Floppy disk support module                         *
//*                                                                            *
//*                   Copyright (C) 2007-2008 Stewart Kay                      *
//******************************************************************************
//
// This module implements functions to access floppy disks and/or disk images
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.4.0 - 18 February 2008, S.J.Kay
// - Added LibDsk providing disk image, floppy access, allowing other
//   drivers to be employed. The existing image and basic Unices floppy
//   access methods remains.
// - Added a side read method to allow disks in LibDsk to correctly read
//   side 1 when the sector header contains side 0. Microbee DSx0 formats
//   can have this problem depending on the format program used.
// - Added disk_format_track() function.
// - Sector range checking is now carried out in this module.
// - Changed itype numbers, 0 is now 'no type selected'.
// - Removed obsolete and unimplemented disk_create() function.
// - Added disk_init function.
//
// v2.3.0 - 21 January 2008, S.J.Kay
// - Added preliminary support for direct floppy access for Unices.
// - Added modio_t structure.
//
// v2.2.0 - 12 December 2007, S.J.Kay
// - added image write protect feature.  A trailing underscore '_' character
//   makes the image write protected.
// - disk->imagerec.datatrack value changed to 1 for non DS80 formats. Not
//   an issue as the system and data tracks have the same sector offsets for
//   either tracks anyway.
//
// v2.1.0 - 24 October 2007, S.J.Kay
// - Added new function disk_read_idfield() to obtain the ID field from the
//   track, side.
// - Undid the changes made in v2.0.0 to the imagerec.datatrack back to how it
//   is in v1.4.0.
// - Microbee DS80 format fixed,  this format has 4 reserved cylinders not 2.
//   disk->imagerec.datatrack = 4
//
// v2.0.0 - 18 October 2007, S.J.Kay
// - DSK format should now be fully complete.
// - Changes to imagerec.datatrack used by all raw and DIP formats now holds
//   the byte offset instead of number of tracks for first data track.
//
// v1.4.0 - 6 October 2007, S.J.Kay
// - Added alternative shortened 3 character filename extentions for raw images.
// - The '.dsk' format now includes 3.5" Modular and should work with other
//   unknown disk formats.  Sector offset information is now read per track.
// - Fixed bug preventing '.DSnn' and '.SSnn' raw images from being opened.
//   Changed image_types[][5] to be image_types[][6].
//
// v1.1.0 - 14 August 2007, S.J.Kay
// - Added Microbee ds40, ss80, ds80, ds82, ds84 raw disk image types.
// - Modified DSK code to detect all Microbee formats.  This needs testing for
//   some formats. (warning in README)
//
// v1.0.0 - 19 June 2007, S.J.Kay
// The following changes/additions have been made:
// - Added new disk image structures,  this now allows for different floppy
//   images to be used instead of only the 40T DS (SBC) images.
// - disk_open function now uses open with write enabled (r+b). This is
//   required to allow disk_write to work.
// - Added an assert to test max_size variable in disk_read function.
// - Implement the disk_write function.
//
// v0.0.0 - 7 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007-2008 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#if USE_LIBDSK
#include <libdsk.h>
#endif

#include "disk.h"
#include "ubee512.h"

//==============================================================================
// structures and variables
//==============================================================================
extern modio_t modio;

#ifdef USE_LIBDSK
//==============================================================================
// Callback function for LibDsk to report messages.
//
//   pass: void
// return: DSK_REPORTFUNC
//==============================================================================
static void disk_libdsk_report (const char *s)
{
 printf("%s\n", s);
}
                
//==============================================================================
// Callback function for LibDsk report end messages.
//
//   pass: void
// return: DSK_REPORTFUNC
//==============================================================================
static void disk_libdsk_report_end (void)
{
}
#endif

//==============================================================================
// Initialise.
//
// Called from fdc.c fdc_init function
//
//   pass: void
// return: int			0 if no errors, else -1
//==============================================================================
int disk_init (void)
{
#ifdef USE_LIBDSK
 dsk_reportfunc_set(disk_libdsk_report, disk_libdsk_report_end);
#endif
 return 0;   
}

//==============================================================================
// Disk open.
//
// Open a disk format for the disk structure requested.  This can one of the
// following:
//
// * original v0.22 raw disk format (40T DS DD) Side0=1st 200K, Side1=2nd 200K
// * CPC-EMU dsk format (ubee512 built-in support)
// * dip format (disk image plus) This is an in-house spec not finalised.
// * ds40 - Microbee 40T DS DD (SBC) S0-T0, S1-T0, S0-T1, S1-T1 ...
// * ss80 - Microbee 80T SS DD (CIAB) S0-T0, S0-T1, S0-T2 ...
// * ds80 - Microbee 80T DS DD (modular) S0-T0, S1-T0, S0-T1, S1-T1 ...
// * ds82 - Microbee 80T DS DD (dream) S0-T0, S1-T0, S0-T1, S1-T1 ...
// * ds84 - Microbee 80T DS DD (pjb) S0-T0, S1-T0, S0-T1, S1-T1 ...
// * LibDsk - Many types
//
//   pass: const char *name
//         disk_t *disk
// return: int				0 if no error, else -1 
//==============================================================================
int disk_open (const char *name, disk_t *disk)
{
 int type_start = 0;

 int itype_temp;
 
 dski_t dski;			// dsk information block
 dskt_t dskt;			// dsk track information block

#ifdef USE_LIBDSK
 dsk_err_t dsk_err;
 dsk_format_t fmt; 
 dsk_cchar_t fdesc;
 dsk_cchar_t fname;
#endif

 char uname[512];

// these formats are for the built in RAW and DSK driver (not LibDsk)
 char const image_types[][6] =
 {
  ".DIP",			// Disk Image Plus (in-house format)
  ".DSK",			// standard CPC DSK format
  ".IMG",			// Nanowasp v0.22 format
  ".NW",			// Nanowasp v0.22 format
  ".DS40",			// raw images
  ".D40",  
  ".SS80",
  ".S80",
  ".DS80",
  ".D80",  
  ".DS82",
  ".D82",  
  ".DS84",
  ".D84",
  ""
 };

// no type selected
 disk->itype = 0;

#ifdef USE_LIBDSK
 if (! strlen(disk->libdsk_format)) // if not using LibDsk
    {
#endif    
// make an upper case copy of the file name  
     int i = 0;
     strcpy(uname, name);
     while (uname[i])
        {
         uname[i] = toupper(uname[i]);
         i++;
        }

// image is write protected if a trailing '_' is found
     disk->wrprot = uname[strlen(uname)-1] == '_';
     if (disk->wrprot)
        uname[strlen(uname)-1] = 0;		//remove the '_'

// direct floppy access. ie /dev/fd0u800/.ds84
#ifdef MINGW
#else
     if (strstr(name, "/dev/") == name)
        {
         char *ch = strstr(name, "/.");
         if (ch)
            *ch = 0;
         type_start = 4;   
        }
#endif 

// open the image or floppy
     disk->fdisk = fopen(name, "r+b");
     if (disk->fdisk == NULL)
        return -1;

// find the disk type
     for (i = type_start; image_types[i][0] != 0; i++)
        {
         if (strstr(uname, image_types[i]) != NULL)
            break;
        }
     if (image_types[i][0] == 0)
        return -1;    

     itype_temp = i + 1;
#ifdef USE_LIBDSK
    }
 else
    {
     disk->wrprot = 0;
     itype_temp = DISK_LIBDSK;
    }    
#endif

// configure parameters for the disk type 
 switch (itype_temp)
    {
     case DISK_DIP    : // DIP (in house spec)
                        fseek(disk->fdisk, -sizeof(disk->imagerec), SEEK_END);
                        fread(&disk->imagerec, sizeof(disk->imagerec), 1, disk->fdisk);
                        if (strcmp(disk->imagerec.id, "DISK-IMAGE-PLUS") != 0)
                           return -1;
                        if ((disk->imagerec.secsize != 128) && (disk->imagerec.secsize != 256) &&
                           (disk->imagerec.secsize != 512) && (disk->imagerec.secsize != 1024))
                           return -1;
                        if (! disk->wrprot)   // if not already '_' write protected
                           disk->wrprot = disk->imagerec.wrprot;
#ifdef use_info_disk_open
                        printf("Image: %s   Vers: %s   Type: %s\n", disk->imagerec.id,
                        disk->imagerec.ver, disk->imagerec.type);
#endif
                        break;
     case DISK_DSK    : // DSK
// get the disk information block
                        fread(&dski, sizeof(dski), 1, disk->fdisk);
                        disk->imagerec.tracks = dski.tracks;
                        disk->imagerec.heads = dski.heads; 

// get the first track information block to get some initial values
                        fread(&dskt, sizeof(dskt), 1, disk->fdisk);
                        disk->imagerec.sectrack = dskt.spt;
                        disk->imagerec.secsize = 128 << dskt.bps;
                        disk->imagerec.systsecofs = dskt.sect_numb;

#ifdef use_info_disk_open
                        printf("Image format is DSK\n");
#endif
                        break;
     case DISK_IMG    : // IMG (nanowasp v0.22 format)
     case DISK_NW     : // NW (nanowasp v0.22 format)
                        disk->imagerec.tracks = 40;
                        disk->imagerec.heads = 2;
                        disk->imagerec.sectrack = 10;
                        disk->imagerec.secsize = 512;           
                        disk->imagerec.datatrack = 1;
                        disk->imagerec.datasecofs = 1;
                        disk->imagerec.systsecofs = 1;

#ifdef use_info_disk_open
                        printf("Image format is RAW 40T DS DD\n");
#endif
                        break;
     case DISK_DS40   : // DS40
     case DISK_D40    : // D40
                        disk->imagerec.tracks = 40;
                        disk->imagerec.heads = 2;
                        disk->imagerec.sectrack = 10;
                        disk->imagerec.secsize = 512;           
                        disk->imagerec.datatrack = 1;
                        disk->imagerec.datasecofs = 1;
                        disk->imagerec.systsecofs = 1;
#ifdef use_info_disk_open
                        printf("Image format is RAW DS40\n");
#endif
                        break;
     case DISK_SS80   : // SS80
     case DISK_S80    : // S80     
                        disk->imagerec.tracks = 80;
                        disk->imagerec.heads = 1;
                        disk->imagerec.sectrack = 10;
                        disk->imagerec.secsize = 512;           
                        disk->imagerec.datatrack = 1;
                        disk->imagerec.datasecofs = 1;
                        disk->imagerec.systsecofs = 1;

#ifdef use_info_disk_open
                       printf("Image format is RAW SS80\n");
#endif
                        break;
     case DISK_DS80   : // DS80 (Microbee modular)
     case DISK_D80    : // D80 (Microbee modular)     
                        disk->imagerec.tracks = 80;
                        disk->imagerec.heads = 2;
                        disk->imagerec.sectrack = 10;
                        disk->imagerec.secsize = 512;           
                        disk->imagerec.datatrack = 2;         // 4 reserved tracks
                        disk->imagerec.datasecofs = 21;       // sectors are 21 to 30
                        disk->imagerec.systsecofs = 1;

#ifdef use_info_disk_open
                       printf("Image format is RAW DS80\n");
#endif
                        break;
     case DISK_DS82   : // DS82
     case DISK_D82    : // D82     
                        disk->imagerec.tracks = 80;
                        disk->imagerec.heads = 2;
                        disk->imagerec.sectrack = 10;
                        disk->imagerec.secsize = 512;           
                        disk->imagerec.datatrack = 1;
                        disk->imagerec.datasecofs = 1;
                        disk->imagerec.systsecofs = 1;
#ifdef use_info_disk_open
                       printf("Image format is RAW DS82\n");
#endif
                        break;
     case DISK_DS84   : // DS84
     case DISK_D84    : // D84     
                        disk->imagerec.tracks = 80;
                        disk->imagerec.heads = 2;
                        disk->imagerec.sectrack = 10;
                        disk->imagerec.secsize = 512;           
                        disk->imagerec.datatrack = 1;
                        disk->imagerec.datasecofs = 1;
                        disk->imagerec.systsecofs = 1;
#ifdef use_info_disk_open
                        printf("Image format is RAW DS84\n");
#endif
                        break;
#ifdef USE_LIBDSK
     case DISK_LIBDSK : // LibDsk
                        disk->itype = 0;
                        // Find the format name
                        int found = 0;
                        fmt = FMT_180K;
                        while ((dg_stdformat(NULL, fmt, &fname, NULL) == DSK_ERR_OK) && (! found))
                           {
                            if (!strcmp(disk->libdsk_format, fname))
                               found = 1;
                            else
                               ++fmt;
                           }

                        if (! found)
                           {   
                            console_output();
                            printf("disk_open: LibDsk format name %s not recognised.\n", disk->libdsk_format);
                            return -1;
                           }
                        // Initialise disk geometry                                                       
                        dsk_err = dg_stdformat(&disk->dg, fmt, NULL, &fdesc);
                        if (dsk_err != DSK_ERR_OK)
                           {
                            console_output();
                            printf("disk_open: LibDsk format name %s not recognised.\n", fdesc);                           
                            return -1;
                           }
                        
                        disk->imagerec.tracks = disk->dg.dg_cylinders;
                        disk->imagerec.heads = disk->dg.dg_heads;
                        disk->imagerec.sectrack = disk->dg.dg_sectors;
                        disk->imagerec.secsize = disk->dg.dg_secsize;
                        disk->imagerec.datatrack = 1;
                        disk->imagerec.datasecofs = disk->dg.dg_secbase;
                        disk->imagerec.systsecofs = disk->dg.dg_secbase;

                        // use side1 as 0 if native formatted disks are being used
                        if ((strcmp(disk->libdsk_format, "ds80") == 0) ||
                            (strcmp(disk->libdsk_format, "ds40") == 0))
                            disk->side1as0 = 1;

                        // Open the disk image or floppy
                        if (strlen(disk->libdsk_type))
                           dsk_err = dsk_open(&disk->self, disk->filename, disk->libdsk_type, NULL);
                        else   
                           dsk_err = dsk_open(&disk->self, disk->filename, NULL, NULL);
                           
                        if (disk->self)
                           dsk_set_retry(disk->self, 5);
                        else
                           {
                            console_output();
                            printf("disk_open: dsk_open error: dsk_err=%d\n", dsk_err);
                            disk->itype = 0;
                            return -1;
                           }
                        break;
#endif
    } 

 disk->itype = itype_temp;

#ifdef use_info_disk_open
 printf("Tracks: %d   Heads: %d   S/T: %d   B/S: %d\n", 
 disk->imagerec.tracks, disk->imagerec.heads,
 disk->imagerec.sectrack, disk->imagerec.secsize);
#endif

 return 0;    
}

//==============================================================================
// Disk close.
//
//   pass: disk_t *disk
// return: void
//==============================================================================
void disk_close (disk_t *disk)
{
#ifdef USE_LIBDSK
 if (disk->itype == DISK_LIBDSK)
    dsk_close(&disk->self);
 else
#endif 
    fclose(disk->fdisk);

 disk->fdisk = NULL;
 disk->itype = 0; 
}

int disk_create(const char *name, disk_t *disk)
{
 return -1;
}

//==============================================================================
// Disk read.
//
//   pass: disk_t *disk
//         char *buf
//         int side
//         int track
//         int sect
//         char rtype			m if a multi sector read operation
// return: int				0 if no errors, else error number
//==============================================================================
int disk_read (disk_t *disk, char *buf, int side, int track, int sect, char rtype)
{
 dskt_t dskt;	// dsk track information block

 int res = 0;

#ifdef USE_LIBDSK
 dsk_err_t dsk_err = 0;
#endif

 int trackofs;
 int sectofs;
 int sectuse;
 int dskofs;

#ifdef use_debug_disk_read_abort
 printf("disk_read: forcing an abort here\n");
 assert(1 != 1);
#endif

#ifdef use_info_disk_read
 printf("Tracks: %d   Heads: %d   S/T: %d   B/S: %d\n", 
 disk->imagerec.tracks, disk->imagerec.heads,
 disk->imagerec.sectrack, disk->imagerec.secsize);
 printf("Track: %3d   Side: %3d   Sector: %3d\n",  track, side, sect);
#endif

 switch (disk->itype)
    {
     case DISK_DIP    : // DIP (in house spec)
                        trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
                        if (track >= disk->imagerec.datatrack)
                           sectuse = sect - disk->imagerec.datasecofs;
                        else
                           sectuse = sect - disk->imagerec.systsecofs;
                        sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
#ifdef use_info_disk_read
                        printf("Image: %s   Vers: %s   Type: %s\n", disk->imagerec.id,
                        disk->imagerec.ver, disk->imagerec.type);
#endif
                        if ((sectuse >= 0) && (sectuse < disk->imagerec.sectrack))
                           {
                            fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);
                            fread(buf, disk->imagerec.secsize, 1, disk->fdisk);
                           }
                        else
                           res = -1;
                        break;
     case DISK_DSK    : // DSK
                        trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
                        dskofs = (((track * disk->imagerec.heads) + side + 1) * 0x100) + 0x100;

// read the track information header for the track
                        sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize;
                        fseek(disk->fdisk, trackofs + sectofs + dskofs - 0x100, SEEK_SET);
                        fread(&dskt, sizeof(dskt), 1, disk->fdisk);
                        disk->imagerec.datasecofs = dskt.sect_numb;
                        disk->imagerec.systsecofs = dskt.sect_numb;
                        sectuse = sect - disk->imagerec.datasecofs;              

// seek and read the sector
                        if ((sectuse >= 0) && (sectuse < disk->imagerec.sectrack))
                           {
                            sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
                            fseek(disk->fdisk, trackofs + sectofs + dskofs, SEEK_SET);
                            fread(buf, disk->imagerec.secsize, 1, disk->fdisk);
                           }
                        else
                           res = -1;
                        break;
     case DISK_IMG    : // IMG (nanowasp v0.22 format)
     case DISK_NW     : // NW (nanowasp v0.22 format)
                        if ((sect > 0) && (sect <= disk->imagerec.sectrack))
                           {
                            fseek(disk->fdisk, ((side * 40 + track) * 10 + (sect - 1)) * 512, SEEK_SET);
                            fread(buf, 512, 1, disk->fdisk);
                           }
                        else
                           res = -1;
                        break;
     case DISK_DS40   : // DS40
     case DISK_D40    : // D40     
     case DISK_SS80   : // SS80
     case DISK_S80    : // S80     
     case DISK_DS80   : // DS80 (Microbee modular)
     case DISK_D80    : // D80 (Microbee modular)     
     case DISK_DS82   : // DS82
     case DISK_D82    : // D82     
     case DISK_DS84   : // DS84
     case DISK_D84    : // D84
                        trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
                        sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize;

                        if (track >= disk->imagerec.datatrack)
                           sectuse = sect - disk->imagerec.datasecofs;
                        else
                           sectuse = sect - disk->imagerec.systsecofs;

                        sectofs += sectuse * disk->imagerec.secsize;

                        if ((sectuse >= 0) && (sectuse < disk->imagerec.sectrack))
                           {
                            fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);
                            fread(buf, disk->imagerec.secsize, 1, disk->fdisk);
                           }
                        else
                           res = -1;
                        break;
#ifdef USE_LIBDSK
     case DISK_LIBDSK : // LibDsk
                        if ((disk->side1as0) && (side == 1))
                           dsk_err = dsk_xread(disk->self, &disk->dg, buf, track, side, track, 0, sect, disk->secsize, NULL);
                        else   
                           dsk_err = dsk_pread(disk->self, &disk->dg, buf, track, side, sect);
                        res =  dsk_err;
                        break;
#endif
    }  

// check if there were any errors and report and exit
 if (res)
    {
#ifdef USE_LIBDSK
     if (disk->itype == DISK_LIBDSK)
        {
         if ((dsk_err != DSK_ERR_NODATA) && (rtype != 'm')) // Don't report multi sector error
            {
             if ((disk->side1as0) && (side == 1))
                printf("disk_read: dsk_xread");
             else   
                printf("disk_read: dsk_pread");
             printf(" error: file=%s dsk_err=%d %s\n", disk->filename, dsk_err, dsk_strerror(dsk_err));
             printf("Track: %3d   Side: %3d   Sector: %3d\n",  track, side, sect);
            } 
        }       
     else
#endif
        { 
         if (rtype != 'm') // Don't report multi sector error
            {
             printf("disk_read: (inbuilt) file=%s res=%d\n", disk->filename, res);
             printf("Track: %3d   Side: %3d   Sector: %3d\n",  track, side, sect);
            } 
        }
     
     return res;
    }
 
 return 0;
}

//==============================================================================
// Disk write.
//
// A flush call is used on each sector write.  This may be moved out at a later
// stage when image (floppy) swapping is implemented.
//
//   pass: disk_t *disk
//         char *buf
//         int side
//         int track
//         int sect
//         char wtype			m if a multi sector read operation
// return: int				0 if no errors, else error number
//==============================================================================
int disk_write (disk_t *disk, char *buf, int side, int track, int sect, char wtype)
{
 dskt_t dskt;			// dsk track information block

 int res = 0;

#ifdef USE_LIBDSK
 dsk_err_t dsk_err = 0;
#endif

 int trackofs;
 int sectofs;
 int sectuse; 
 int dskofs;

#ifdef use_info_disk_read
 printf("Tracks: %d   Heads: %d   S/T: %d   B/S: %d\n", 
 disk->imagerec.tracks, disk->imagerec.heads,
 disk->imagerec.sectrack, disk->imagerec.secsize);
 printf("Track: %3d   Side: %3d   Sector: %3d\n",  track, side, sect);
#endif

 switch (disk->itype)
    {
     case DISK_DIP    : // DIP (in house spec)
                        trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
                        if (track >= disk->imagerec.datatrack)
                           sectuse = sect - disk->imagerec.datasecofs;
                        else
                           sectuse = sect - disk->imagerec.systsecofs;       
                        sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
#ifdef use_info_disk_read
                        printf("Image: %s   Vers: %s   Type: %s\n", disk->imagerec.id,
                        disk->imagerec.ver, disk->imagerec.type);
#endif
                        if ((sectuse >= 0) && (sectuse < disk->imagerec.sectrack))
                           {
                            fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);
                            fwrite(buf, disk->imagerec.secsize, 1, disk->fdisk);
                            fflush(disk->fdisk);
                           }
                        else
                           res = -1;
                        break;
     case DISK_DSK    : // DSK
                        trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
                        dskofs = (((track * disk->imagerec.heads) + side + 1) * 0x100) + 0x100;

// read the track information header for the track
                        sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize;
                        fseek(disk->fdisk, trackofs + sectofs + dskofs - 0x100, SEEK_SET);
                        fread(&dskt, sizeof(dskt), 1, disk->fdisk);
                        disk->imagerec.datasecofs = dskt.sect_numb;
                        disk->imagerec.systsecofs = dskt.sect_numb;
                        sectuse = sect - disk->imagerec.datasecofs;

// seek and write the sector
                        sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
                        if ((sectuse >= 0) && (sectuse < disk->imagerec.sectrack))
                           {
                            fseek(disk->fdisk, trackofs + sectofs + dskofs, SEEK_SET);
                            fwrite(buf, disk->imagerec.secsize, 1, disk->fdisk);
                            fflush(disk->fdisk);
                           } 
                        else
                           res = -1;
                        break;
     case DISK_IMG    : // IMG (nanowasp v0.22 format)
     case DISK_NW     : // NW (nanowasp v0.22 format)
                        if ((sect > 0) && (sect <= disk->imagerec.sectrack))
                           {
                            fseek(disk->fdisk, ((side * 40 + track) * 10 + (sect - 1)) * 512, SEEK_SET);
                            fwrite(buf, 512, 1, disk->fdisk);
                            fflush(disk->fdisk);
                           }
                        else
                           res = -1;
                        break;
     case DISK_DS40   : // DS40
     case DISK_D40    : // D40     
     case DISK_SS80   : // SS80
     case DISK_S80    : // S80     
     case DISK_DS80   : // DS80 (Microbee modular)
     case DISK_D80    : // D80 (Microbee modular)     
     case DISK_DS82   : // DS82
     case DISK_D82    : // D82     
     case DISK_DS84   : // DS84
     case DISK_D84    : // D84
                        trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
                        sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize;

                        if (track >= disk->imagerec.datatrack)
                           sectuse = sect - disk->imagerec.datasecofs;
                        else
                           sectuse = sect - disk->imagerec.systsecofs;

                        sectofs += sectuse * disk->imagerec.secsize;

                        if ((sectuse >= 0) && (sectuse < disk->imagerec.sectrack))
                           {
                            fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);
                            fwrite(buf, disk->imagerec.secsize, 1, disk->fdisk);
                            fflush(disk->fdisk);
                           }
                        else
                           res = -1;
                        break;
#ifdef USE_LIBDSK
     case DISK_LIBDSK : // LibDsk
                        if ((disk->side1as0) && (side == 1))
                           dsk_err = dsk_xwrite(disk->self, &disk->dg, buf, track, side, track, 0, sect, disk->secsize, (int)NULL);
                        else
                           dsk_err = dsk_pwrite(disk->self, &disk->dg, buf, track, side, sect);
                        res = dsk_err;
                        break;
#endif                        
    }  

// check if there were any errors and report and exit
 if (res)
    {
#ifdef USE_LIBDSK
     if (disk->itype == DISK_LIBDSK)
        {
         if ((dsk_err != DSK_ERR_NODATA) && (wtype != 'm')) // Don't report multi sector error
            {
             if ((disk->side1as0) && (side == 1))
                printf("disk_write: dsk_xwrite");
             else   
                printf("disk_write: dsk_pwrite");
             printf(" error: file=%s dsk_err=%d %s\n", disk->filename, dsk_err, dsk_strerror(dsk_err));
             printf("Track: %3d   Side: %3d   Sector: %3d\n",  track, side, sect);
            } 
        }       
     else
#endif
        { 
         if (wtype != 'm') // Don't report multi sector error
            {
             printf("disk_write: (inbuilt) file=%s res=%d\n", disk->filename, res);
             printf("Track: %3d   Side: %3d   Sector: %3d\n",  track, side, sect);
            } 
        }
     
     return res;
    }
 
 return 0;
}

//==============================================================================
// Disk read ID field.
//
// Fills the read_addr_t structure id with the ID field.  For internal drivers
// the structure is set to the passed side and track values and values from
// the disk->imagerec structure.
//
//   pass: disk_t *disk
//         read_addr_t *idfield
//         int side
//         int track
// return: int				0 if no errors, else error number
//==============================================================================
int disk_read_idfield (disk_t *disk, read_addr_t *idfield, int side, int track)
{
 dskt_t dskt;			// dsk track information block
 
 int res = 0; 

#ifdef USE_LIBDSK
 dsk_err_t dsk_err;
 DSK_FORMAT result;
#endif

 int trackofs;
 int sectofs;
 int dskofs;

 switch (disk->itype)
    {
     case DISK_DSK    : // if DSK format (not LibDsk)
                        trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
                        dskofs = (((track * disk->imagerec.heads) + side + 1) * 0x100) + 0x100;

                        // read the track information header for the track
                        sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize;
                        fseek(disk->fdisk, trackofs + sectofs + dskofs - 0x100, SEEK_SET);
                        fread(&dskt, sizeof(dskt), 1, disk->fdisk);

                        idfield->track = track;
                        idfield->side = side;
                        idfield->seclen = disk->imagerec.secsize >> 8; 
                        idfield->secaddr = dskt.sect_numb;
                        break;
#ifdef USE_LIBDSK
     case DISK_LIBDSK : // LibDsk
                        dsk_err = dsk_psecid(disk->self, &disk->dg, track, side, &result);
                        idfield->track = result.fmt_cylinder;
                        idfield->side = result.fmt_head;
                        idfield->seclen = result.fmt_secsize >> 8; 
                        idfield->secaddr = result.fmt_sector;

                        if (dsk_err)
                           printf("disk_read_idfield: dsk_psecid error: file=%s dsk_err=%d %s\n", disk->filename, dsk_err, dsk_strerror(dsk_err));
                        res = dsk_err;
                        break;
#endif
              default : // internal raw and DIP formats
                        idfield->track = track;
                        idfield->side = side;
                        idfield->seclen = disk->imagerec.secsize >> 8; 
                        if (track >= disk->imagerec.datatrack)
                           idfield->secaddr = disk->imagerec.datasecofs;
                        else
                           idfield->secaddr = disk->imagerec.systsecofs;
                        break;
    }

 idfield->crc1 = 0xff;
 idfield->crc2 = 0xff;

 return res;
}

//==============================================================================
// Disk write protected query.
//
//   pass: disk_t *disk
// return: int				write protect status
//==============================================================================
int disk_iswrprot (disk_t *disk)
{
#ifdef USE_LIBDSK
 dsk_err_t dsk_err;
 unsigned char result;
#endif

 switch (disk->itype)
    {
#ifdef USE_LIBDSK
     case DISK_LIBDSK : // LibDsk
                        dsk_err = dsk_drive_status(disk->self, &disk->dg, 0, &result);
                        disk->wrprot = (result & DSK_ST3_RO);

                        if (dsk_err)
                           printf("disk_iswrprot: dsk_drive_status error: file=%s dsk_err=%d %s\n", disk->filename, dsk_err, dsk_strerror(dsk_err));
                        break;

                        break;
              default : // internal formats
                        break;               
#endif                        
    }

 return disk->wrprot;
}

//==============================================================================
// Disk set geometry member.
//
//   pass: disk_t *disk
// return: void
//==============================================================================
int disk_setgeom_member (disk_t *disk)
{
 return 0;
}

//==============================================================================
// Disk format track
//
//   pass: disk_t *disk
//         int side
//         char *buf
//         int ddense			MFM=1, FM=0
//         int track
//         int side
//         int sectors
// return: int				0 if no error
//==============================================================================
int disk_format_track (disk_t *disk, char *buf, int ddense, int track, int side, int sectors)
{
 int res;
 
#ifdef USE_LIBDSK
 dsk_err_t dsk_err;
 static DSK_GEOMETRY geom;
 static DSK_FORMAT format[128]; // up to 128 sector headers
 int i;
 int s;
#endif

 switch (disk->itype)
    {
#ifdef USE_LIBDSK
     case DISK_LIBDSK : // LibDsk
                        memcpy(&geom, &disk->dg, sizeof(geom));
                        geom.dg_fm = (ddense == 0);
                        i = 0;
                        for (s = 0; s < sectors; s++)
                           {
                            format[s].fmt_cylinder = buf[i++];
                            format[s].fmt_head = buf[i++];
                            format[s].fmt_sector = buf[i++];
                            format[s].fmt_secsize = buf[i++] << 8;
                           }
                        dsk_err = dsk_pformat(disk->self, &geom, track, side, format, 0xe5);
                        
                        if (dsk_err)
                           printf("disk_format_track: dsk_pformat error: file=%s dsk_err=%d %s\n", disk->filename, dsk_err, dsk_strerror(dsk_err));
                        res = dsk_err;
                        break;
#endif                        
              default : // internal formats
                        res = 0;
                        break;
    }

 return res;                       
}
