/* Floppy Disk Support Header */

#ifndef HEADER_DISK_H
#define HEADER_DISK_H

#include <inttypes.h>

#if USE_LIBDSK
#include <libdsk.h>
#endif

#include "ubee512.h"

// Disk image types
#define DISK_DIP	1
#define DISK_DSK	2
#define DISK_IMG	3
#define DISK_NW		4
#define DISK_DS40	5
#define DISK_D40	6
#define DISK_SS80	7
#define DISK_S80	8
#define DISK_DS80	9
#define DISK_D80	10
#define DISK_DS82	11
#define DISK_D82	12
#define DISK_DS84	13
#define DISK_D84	14
#define DISK_LIBDSK	15

// Disk image record information (adjust filler for 512 bytes)
typedef struct diski_t
{
 char id[32];		// disk image ID
 char ver[16];		// disk image version
 char type[32];		// disk image type

 uint16_t wrprot;	// write protected
 uint16_t tracks;	// number of tracks
 uint16_t heads;	// number of heads (sides)
 uint16_t secsize;	// bytes per sector
 uint16_t sectrack;	// sectors per track
 uint16_t datatrack;	// start of data tracks (after system)
 uint16_t systsecofs;	// offset to apply to system sectors
 uint16_t datasecofs;	// offset to apply to data sectors
 uint8_t skewsa[3];	// skew algorithm info for system tracks
 uint8_t skewda[3];	// skew algorithm info for data tracks
 uint8_t skewsd[128]; 	// skew data for system tracks
 uint8_t skewdd[128]; 	// skew data for data tracks
 uint16_t secterrs;	// size of sector error table (0 if not used) 
 char fill[152];	// filler to make structure 512 bytes long
}diski_t;

typedef struct disk_t
{
 FILE *fdisk;
 int wrprot;
 int itype;		// image type
 int probesec;		// probe sector
 int secsize;		// sector size
 diski_t imagerec;
 char filename[512];
#ifdef USE_LIBDSK
 int side1as0;
 char libdsk_type[40];
 char libdsk_format[40];
 DSK_PDRIVER self;
 DSK_GEOMETRY dg;
#endif
}disk_t;

// dsk disk structure, first 0x100 bytes of image is header the disk
// information block.
typedef struct dski_t
{
 char cpcemu[0x22];		// 0x00 - 0x21
 char unused1[0x0e];		// 0x22 - 0x2f
 uint8_t tracks;		// 0x30
 uint8_t heads;			// 0x31
 uint8_t size_one_ta;		// 0x32
 uint8_t size_one_tb;		// 0x33 
 char unused2[0x100-0x34];      // 0x34 - 0xff
}dski_t;

// dsk disk structure, track information block. We use this to figure out
// the SPT and BPS for all other tracks on the disk.
typedef struct dskt_t
{
 char track_info[0x0d];		// 0x00 - 0x0c
 char unused3[0x03];            // 0x0d - 0x0f
 uint8_t tracks_n;		// 0x10
 uint8_t heads_n;		// 0x11
 char unused4[0x02];		// 0x12 - 0x13
 uint8_t bps;			// 0x14
 uint8_t spt;			// 0x15
 uint8_t gap3;			// 0x16
 uint8_t filling_byte;		// 0x17
 uint8_t track_numb;		// 0x18
 uint8_t hnumb_sectid; 		// 0x19
 uint8_t sect_numb;		// 0x1a
 uint8_t bps_x;			// 0x1b
 uint8_t state1_errcode; 	// 0x1c
 uint8_t state2_errcode;  	// 0x1d
 char unused5[0x02];		// 0x1e - 0x1f
 char unused6[0x100-0x20];	// 0x20 - 0xff
}dskt_t;

typedef struct read_addr_t
{
 uint8_t track;
 uint8_t side;
 uint8_t secaddr;
 uint8_t seclen;
 uint8_t crc1;
 uint8_t crc2;
}read_addr_t;

int disk_init (void);
int disk_open (const char *name, disk_t *disk);
void disk_close (disk_t *disk);
int disk_read (disk_t *disk, char *buf, int side, int track, int sect, char rtype);
int disk_write (disk_t *disk, char *buf, int side, int track, int sect, char wtype);
int disk_read_idfield (disk_t *disk, read_addr_t *idfield, int side, int track);
int disk_format_track(disk_t *disk, char *buf, int ddense, int track, int side, int sectors);
int disk_iswrprot (disk_t *disk);

#endif  /* HEADER_DISK_H */
