//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                              WD2793 FDC module                             *
//*                                                                            *
//*                   Copyright (C) 2007-2008 Stewart Kay                      *
//******************************************************************************
//
// Emulates the WD2793 floppy disk controller.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.4.0 - 13 February 2008, S.J.Kay
// - Make changes to incorporate LibDsk and multio reads.  Sector range
//   checking is now hanlded by the disk.c module.
// - Added emulation of the FDC write track command.  Format and Init
//   programs now work.
// - Changes to the fdc_data_w() function to handle multi sector writes.
// - Added sector size probing option for read/write sectors.
// - Added error checking and setting of status registers has been revised.
// - Sector size probing option to allow protected disks to work.
// - Added a disk drive activity status function call.
// - Removed malloc and use static intances of fdc_drive[drive].disk
// - Restructured the disk_t to inlude file names, etc
//
// v2.3.0 - 21 January 2008, S.J.Kay
// - Print references to images now includes floppy disk.
// - Added modio_t structure.
//
// v2.2.0 - 12 December 2007, S.J.Kay
// - Added disk write protection to FDC_WRITESECT command.
// - The FDC_READSECT and FDC_WRITESECT commands did not have the correct
//   sectoffs value for DS80 DSK format images when the data track sector
//   offset was obtained.  The problem only affected DSK images where the
//   first sector number is bigger than 1. Now use the disk_read_idfield()
//   function to obtain the sectoffs value.
// - No loading of the fallback boot disk if emulating a machine of less than
//   128K memory.
//
// v2.1.0 - 24 October 2007, S.J.Kay
// - The FDC_READADDR command now calls a new disk_read_idfield() function to
//   obtain the track ID field.  The previous code had assumed all sectors were
//   in the 1 to 10 range and would not work on the 256TC and the DS80 Modular
//   disk images,  the sector size also returns the size for the disk selected.
// - Changes made to allow disk formats that have sectors that start from
//   greater than 1 to be read. i.e DS80 Microbee disks are 10 sectors, with
//   sector numbers 21-30 on the data tracks.
// - Implement the modelx information structure.
//
// v2.0.0 - 7 October 2007, S.J.Kay
// - Each model emulated now has a preferred boot image.  If that image can not
//   be opened then the fallback image is used.
// - Added fdc_status0 variable for a --nodisk option.  This returns a status
//   value of 0 from the fdc_ext_r() function, and is used to simulate a no
//   disk in drive condition.
//
// v1.4.0 - 26 September 2007, S.J.Kay
// - Changes to ignore functions if a ROM model is being emulated.
// - Added changes to error reporting, replaced exit(1) with return -1 in
//   fdc_loaddisk() function.  
//
// v1.3.0 - 1 September 2007, S.J.Kay
// - Changes some error messages and redirection.
//
// v1.0.0 - 2 July 2007, S.J.Kay
// - Implement DSK disk image format instead of raw images.
// - All references to '10' sectors per track have been replaced by
//   drive->disk->imagerec.sectrack and references to '512' have been replaced
//   by drive->disk->imagerec.secsize.
// - Disk images are no longer assumed to MAXTRACK=39,  for now change to
//   255 (largest value allowed for WD2793 ?)
// - BUFSSZE increased from 10x512 to 10x1024 byte sectors.
// - Added 2 more drives (now is A-D).  The drive images can be specified on
//   the command line. If drive 'A' is not specified then "boot.dsk" will be
//   used by default.  Drives 'B-D' must be specified if required.
// - implemented case FDC_WRITESECT: in fdc_cmd function.
// - implemented fdc_data_w function.
// - in fdc_data_r changed the ctrl_status value at end sector read.  The
//   value of FDC_WRPROT did not look correct here.
// - Added a second drive.  
//
// v0.0.0 - 5 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007-2008 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>

#include "fdc.h"
#include "disk.h"
#include "ubee512.h"

static int fdc_loaddisk (char *name, int drive, int report);
static void fdc_unloaddisk (int drive);
static int fdc_bootimage (void);

//==============================================================================
// structures and variables
//==============================================================================
fdc_drive_t fdc_drive[FDC_NUMDRIVES];

static int ctrl_side, ctrl_drive, ctrl_ddense,
           ctrl_rdata, ctrl_rtrack, ctrl_rsect,
           ctrl_intrq, ctrl_drq, ctrl_status,
           ctrl_stepdir;

static int fdc_error;

static char buf[FDC_BUFSIZE];

static int cmdx;
static int bytes_left;
static int bytes_sect;
static int buf_index;
static int write_track_action;
static int sector_header_pos;
static int sector_count;
static int data_same;
static int data_last;
static int data_columns;

int fdc_status0;

extern int model;
extern model_t modelx;

extern char userhome_diskpath[];
extern char *model_args[];

extern modio_t modio;

//==============================================================================
// Initialise
//
//   pass: void
// return: int			0 if no errors, else -1
//==============================================================================
int fdc_init (void)
{
 int i;
 int res;

 if (modelx.rom)
    return 0;

 disk_init();

 for (i = 0; i < FDC_NUMDRIVES; i++)
    {
     fdc_drive[i].disk.fdisk = NULL;
     fdc_drive[i].disk.itype = 0;
    } 

 res = fdc_bootimage();
 
 if (strlen(fdc_drive[1].disk.filename) && (res == 0))
    {
     printf("Image B: %s\n", fdc_drive[1].disk.filename);
     res = (fdc_loaddisk(fdc_drive[1].disk.filename, 1, 1));
    }    

 if (strlen(fdc_drive[2].disk.filename) && (res == 0))
    {
     printf("Image C: %s\n", fdc_drive[2].disk.filename);
     res = (fdc_loaddisk(fdc_drive[2].disk.filename, 2, 1));
    }    
     
 if (strlen(fdc_drive[3].disk.filename) && (res == 0))
    {
     printf("Image D: %s\n", fdc_drive[3].disk.filename);
     res = (fdc_loaddisk(fdc_drive[3].disk.filename, 3, 1));
    }

 cmdx = -1;

 if (res < 0)
    return -1;
 return 0;   
}

//==============================================================================
// De-initialise
//
//   pass: void
// return: int			0
//==============================================================================
int fdc_deinit (void)
{
 int i;

 if (modelx.rom)
    return 0;

 for (i = 0; i < FDC_NUMDRIVES; i++)
    fdc_unloaddisk(i);
 
 return 0;
}

//==============================================================================
// reset the controller
//
//   pass: void
// return: int			0
//==============================================================================
int fdc_reset (void)
{
 int i;
 
 if (modelx.rom)
    return 0;

 ctrl_side = ctrl_drive = ctrl_ddense = 0;
 ctrl_rtrack = ctrl_intrq = ctrl_drq = 0;
 ctrl_stepdir = +1;

 ctrl_status = FDC_TRACK0;

 for (i = 0; i < FDC_NUMDRIVES; i++)
     fdc_drive[i].track = 0;

// if (fdc_drive[ctrl_drive].disk.itype == 0)
//    ctrl_status |= FDC_INDEXPULSE;              /* set index pulse if no disk */

// if (disk_iswrprot(&fdc_drive[ctrl_drive].disk)) 
//    ctrl_status |= FDC_WRPROT;
 
 bytes_left = 0;
 cmdx = -1;

 return 0;
}

//==============================================================================
// load disk
//
//   pass: const char *name
//         int drive
//         int report		1 if require error message if file not opened
// return: int			0 if no errors,  -1 if malloc error, else -2
//==============================================================================
static int fdc_loaddisk (char *name, int drive, int report)
{
 fdc_unloaddisk(drive);

 if (disk_open(name, &fdc_drive[drive].disk) != 0)
    {
     fdc_unloaddisk(drive);
     if (report)
        {
         console_output();
         printf("fdc_loaddisk: Unable to open %s\n", name);
         printf("fdc_loaddisk: Check image/floppy disk is not set as read only!\n");
        }
     return -2;
    }

 return 0;
}

//==============================================================================
// un-load disk
//
//   pass: int drive		drive number
// return: void
//==============================================================================
static void fdc_unloaddisk (int drive)
{
 if (fdc_drive[drive].disk.itype)
    {
     disk_close(&fdc_drive[drive].disk);
     fdc_drive[drive].disk.fdisk = NULL;
     fdc_drive[drive].disk.itype = 0;
    } 
}

//==============================================================================
// Load boot image.
//
// Load the default boot disk for the model being emulated.  if the default
// boot image is not found then the fall-back image will be booted.
//
//   pass: 
// return: int			0 if no errors,  -1 if malloc error, else -2
//==============================================================================
static int fdc_bootimage (void)
{
 char name[512];
 int res;
 
// if an image for drive A: was specified then load that one
 if (strlen(fdc_drive[0].disk.filename))
    {
     printf("Image A: %s\n", fdc_drive[0].disk.filename);
     res = (fdc_loaddisk(fdc_drive[0].disk.filename, 0, 1));     
    }    
 else
// load the default image for the model being emulated
    {   
     snprintf(name, sizeof(name), "%s%s.dsk", userhome_diskpath, model_args[model]);
     printf("Image A: %s\n", name);
     res = (fdc_loaddisk(name, 0, 0));

// if the default image was not found then use the fallback boot image
     if (res == -2)
        {
         if (modelx.ram < 128)
            {
             console_output();
             printf("Image not able to be opened, aborting...\n");
            } 
         else
            {   
             snprintf(name, sizeof(name), "%s"BOOT_IMAGE, userhome_diskpath);
             printf("Image not able to be opened, trying fall-back boot image\n");
             printf("Image A: %s\n", name);
             res = (fdc_loaddisk(name, 0, 1));
            } 
        }
    }

 return res;
}

//==============================================================================
// FDC command
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void fdc_cmd (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 int cmd;

 read_addr_t *idfield = (read_addr_t *)buf;
 
 if (modelx.rom)
    return;

// clear this in case command aborts (ie for any bad floppy sector reads)
 bytes_left = 0;

 if (fdc_drive[ctrl_drive].disk.itype == 0)
    {
     ctrl_status = FDC_RECNOTFOUND;
     return;
    } 

 ctrl_intrq = 0;  // new command clears interrupt, FIXME

 cmd = (data >> 4) & 0x0F;  // get command bits
 if ( ((cmd >> 1) >= 1) && ((cmd >> 1) <= 5) )
    cmd &= 0x0E;  // get rid of data bits in the command field

 cmdx = cmd;

 switch (cmd)
    {
//------------------------------------------------------------------------------
// Restore - Seek track 0.
//------------------------------------------------------------------------------
     case FDC_RESTORE:
        if (modio.fdc)
           {
            printf("fdc_cmd: restore\n");
            if (modio.level)
               fprintf(modio.log, "fdc_cmd: restore\n");
           }

        fdc_drive[ctrl_drive].track = ctrl_rtrack = 0;

        ctrl_intrq = 1;
        ctrl_status = FDC_TRACK0;
        if (data & (FDC_LOADHEAD | FDC_VERIFY))
           ctrl_status |= FDC_HEADLOADED;
        if (disk_iswrprot(&fdc_drive[ctrl_drive].disk))
           ctrl_status |= FDC_WRPROT;
        if (fdc_drive[ctrl_drive].disk.itype == 0)
           ctrl_status |= FDC_INDEXPULSE;
        break;

//------------------------------------------------------------------------------
// Seek.
//------------------------------------------------------------------------------
     case FDC_SEEK:
        if (modio.fdc)
           {
            printf("fdc_cmd: seek\n");
            if (modio.level)
               fprintf(modio.log, "fdc_cmd: seek\n");
           }

        fdc_drive[ctrl_drive].track += ctrl_rdata - ctrl_rtrack;
        if (fdc_drive[ctrl_drive].track < 0)
           fdc_drive[ctrl_drive].track = 0;
        else
           if (fdc_drive[ctrl_drive].track > FDC_MAXTRACK)
              fdc_drive[ctrl_drive].track = FDC_MAXTRACK;

        if (ctrl_rdata < ctrl_rtrack)
           ctrl_stepdir = -1;
        else
           ctrl_stepdir = +1;
        ctrl_rtrack = ctrl_rdata;

        ctrl_intrq = 1;
        ctrl_status = 0;
        if (fdc_drive[ctrl_drive].track == 0)
           ctrl_status |= FDC_TRACK0;
        if (data & (FDC_LOADHEAD | FDC_VERIFY))
           ctrl_status |= FDC_HEADLOADED;
        if (disk_iswrprot(&fdc_drive[ctrl_drive].disk))
           ctrl_status |= FDC_WRPROT;
        if (fdc_drive[ctrl_drive].disk.itype == 0)
           ctrl_status |= FDC_INDEXPULSE;
        break;

//------------------------------------------------------------------------------
// Step - Steps the same direction as the last step-x command.
//------------------------------------------------------------------------------
     case FDC_STEP:
        if (modio.fdc)
           {
            printf("fdc_cmd: step\n");
            if (modio.level)
               fprintf(modio.log, "fdc_cmd: step\n");
           }
           
        fdc_drive[ctrl_drive].track += ctrl_stepdir;
        if (fdc_drive[ctrl_drive].track < 0)
           fdc_drive[ctrl_drive].track = 0;
        else
           if (fdc_drive[ctrl_drive].track > FDC_MAXTRACK)
              fdc_drive[ctrl_drive].track = FDC_MAXTRACK;

        if (data & FDC_UPDATETRACK)
           ctrl_rtrack += ctrl_stepdir;

        ctrl_intrq = 1;
        ctrl_status = 0;
        if (fdc_drive[ctrl_drive].track == 0)
           ctrl_status |= FDC_TRACK0;
        if (data & (FDC_LOADHEAD | FDC_VERIFY))
           ctrl_status |= FDC_HEADLOADED;
        if (disk_iswrprot(&fdc_drive[ctrl_drive].disk))
           ctrl_status |= FDC_WRPROT;
        if (fdc_drive[ctrl_drive].disk.itype == 0)
           ctrl_status |= FDC_INDEXPULSE;
        break;

//------------------------------------------------------------------------------
// Step-in - Step in towards centre of disk.
//------------------------------------------------------------------------------
     case FDC_STEPIN:
        if (modio.fdc)
           {
            printf("fdc_cmd: stepin\n");
            if (modio.level)
               fprintf(modio.log, "fdc_cmd: stepin\n");
           }
        fdc_drive[ctrl_drive].track++;
        if (fdc_drive[ctrl_drive].track > FDC_MAXTRACK)
           fdc_drive[ctrl_drive].track = FDC_MAXTRACK;

        if (data & FDC_UPDATETRACK)
           ctrl_rtrack++;

        ctrl_stepdir = +1;

        ctrl_intrq = 1;
        ctrl_status = 0;
        if (fdc_drive[ctrl_drive].track == 0)
           ctrl_status |= FDC_TRACK0;
        if (data & (FDC_LOADHEAD | FDC_VERIFY))
           ctrl_status |= FDC_HEADLOADED;
        if (disk_iswrprot(&fdc_drive[ctrl_drive].disk))
           ctrl_status |= FDC_WRPROT;
        if (fdc_drive[ctrl_drive].disk.itype == 0)
           ctrl_status |= FDC_INDEXPULSE;
        break;

//------------------------------------------------------------------------------
// Step-out - Step out towards track 0.
//------------------------------------------------------------------------------
     case FDC_STEPOUT:
        if (modio.fdc)
           {
            printf("fdc_cmd: stepout\n");
            if (modio.level)
               fprintf(modio.log, "fdc_cmd: stepout\n");
           }

        fdc_drive[ctrl_drive].track--;
        if (fdc_drive[ctrl_drive].track < 0)
           fdc_drive[ctrl_drive].track = 0;

        if (data & FDC_UPDATETRACK)
           ctrl_rtrack--;

        ctrl_stepdir = -1;

        ctrl_intrq = 1;
        ctrl_status = 0;
        if (fdc_drive[ctrl_drive].track == 0)
           ctrl_status |= FDC_TRACK0;
        if (data & (FDC_LOADHEAD | FDC_VERIFY))
           ctrl_status |= FDC_HEADLOADED;
        if (disk_iswrprot(&fdc_drive[ctrl_drive].disk))
           ctrl_status |= FDC_WRPROT;
        if (fdc_drive[ctrl_drive].disk.itype == 0)
           ctrl_status |= FDC_INDEXPULSE;
        break;

//------------------------------------------------------------------------------
// Read sector - Read single and multi sectors.
//
// A probing option can be used to sniff the track being read to determine
// the sector size to read,  this may be required for special formats.
//------------------------------------------------------------------------------
     case FDC_READSECT:
        if (modio.fdc)
           {
            printf("fdc_cmd: readsect\n");
            if (modio.level)
               fprintf(modio.log, "fdc_cmd: readsect\n");
           }

// find the sector size
        if (fdc_drive[ctrl_drive].disk.probesec)  // if probe for it
           {
            disk_read_idfield(&fdc_drive[ctrl_drive].disk, idfield, ctrl_side, fdc_drive[ctrl_drive].track);
            fdc_drive[ctrl_drive].disk.secsize = idfield->seclen << 8;
            fdc_drive[ctrl_drive].disk.dg.dg_secsize = fdc_drive[ctrl_drive].disk.secsize;
           }
        else
           fdc_drive[ctrl_drive].disk.secsize = fdc_drive[ctrl_drive].disk.imagerec.secsize;
           
// read 1 sector        
        fdc_error = disk_read(&fdc_drive[ctrl_drive].disk, buf, ctrl_side, fdc_drive[ctrl_drive].track, ctrl_rsect, 0);
        buf_index = 0;
        if (! fdc_error)
           bytes_left = fdc_drive[ctrl_drive].disk.secsize;

// multisector read
        if (data & FDC_MULTISECT)
           {
            if (! fdc_error)
               do 
                  {
                   ctrl_rsect++;
                   fdc_error = disk_read(&fdc_drive[ctrl_drive].disk, &buf[bytes_left], ctrl_side, fdc_drive[ctrl_drive].track, ctrl_rsect, 'm');
                   if (! fdc_error)
                      bytes_left += fdc_drive[ctrl_drive].disk.secsize;
                  }
               while (! fdc_error);
           }     

// check errors and set status register
        ctrl_status = 0;
        
        if (fdc_error)
           {
#ifdef USE_LIBDSK
            if (fdc_drive[ctrl_drive].disk.itype == DISK_LIBDSK)
               {
                switch (fdc_error)
                   {
                    case  DSK_ERR_NOTRDY : ctrl_status = FDC_NOTREADY;
                                           break;
                    case  DSK_ERR_NODATA : ctrl_status = FDC_RECNOTFOUND;
                                           break;
                    case DSK_ERR_DATAERR : ctrl_status = FDC_CRCERROR;
                                           break;
                    case  DSK_ERR_NOADDR : ctrl_status = FDC_RECNOTFOUND;
                                           break;
                    case  DSK_ERR_SYSERR : ctrl_status = FDC_RECNOTFOUND;
                                           break;
                                 default : ctrl_status = FDC_RECNOTFOUND;
                                           break;
                   }
               }
            else    
#endif
               ctrl_status = FDC_RECNOTFOUND;
           }

        ctrl_status |= FDC_BUSY | FDC_DRQ;
        ctrl_drq = 1;
        break;
//------------------------------------------------------------------------------
// Write sector - Write single and multi sectors.
//
// A probing option can be used to sniff the track being written to determine
// the sector size to write,  this may be required for special formats.
//------------------------------------------------------------------------------
     case FDC_WRITESECT:
        if (modio.fdc)
           {
            printf("fdc_cmd: writesect\n");
            if (modio.level)
               fprintf(modio.log, "fdc_cmd: writesect\n");
           }

// find the sector size
        if (fdc_drive[ctrl_drive].disk.probesec)  // if probe for it
           {
            disk_read_idfield(&fdc_drive[ctrl_drive].disk, idfield, ctrl_side, fdc_drive[ctrl_drive].track);
            fdc_drive[ctrl_drive].disk.secsize = idfield->seclen << 8;
            fdc_drive[ctrl_drive].disk.dg.dg_secsize = fdc_drive[ctrl_drive].disk.secsize;
           }
        else
           fdc_drive[ctrl_drive].disk.secsize = fdc_drive[ctrl_drive].disk.imagerec.secsize;
           
        buf_index = 0;
        bytes_sect = 0;
        bytes_left = fdc_drive[ctrl_drive].disk.secsize;

        if (data & FDC_MULTISECT)
           bytes_left = sizeof(buf);

        ctrl_status = FDC_BUSY | FDC_DRQ;
        if (disk_iswrprot(&fdc_drive[ctrl_drive].disk))
           {
            ctrl_status |= FDC_WRPROT;
            break;
           }

        ctrl_drq = 1;
        break;

//------------------------------------------------------------------------------
// Read address - Read next ID field.
//------------------------------------------------------------------------------
     case FDC_READADDR:
        if (modio.fdc)
           {
            printf("fdc_cmd: readaddr\n");
            if (modio.level)
               fprintf(modio.log, "fdc_cmd: readaddr\n");
           }

        bytes_left = 6;
        buf_index = 0;
        fdc_error = disk_read_idfield(&fdc_drive[ctrl_drive].disk, idfield, ctrl_side, fdc_drive[ctrl_drive].track);
        ctrl_rsect = idfield->track;

// check errors and set status register
        ctrl_status = 0;
        
        if (fdc_error)
           {
#ifdef USE_LIBDSK
            if (fdc_drive[ctrl_drive].disk.itype == DISK_LIBDSK)
               {
                switch (fdc_error)
                   {
                    case  DSK_ERR_NOTRDY : ctrl_status = FDC_NOTREADY;
                                           break;
                    case  DSK_ERR_NODATA : ctrl_status = FDC_RECNOTFOUND;
                                           break;
                    case DSK_ERR_DATAERR : ctrl_status = FDC_CRCERROR;
                                           break;
                    case  DSK_ERR_NOADDR : ctrl_status = FDC_RECNOTFOUND;
                                           break;
                    case  DSK_ERR_SYSERR : ctrl_status = FDC_RECNOTFOUND;
                                           break;
                                 default : ctrl_status = FDC_RECNOTFOUND;
                                           break;
                   }
               }
            else    
#endif
               ctrl_status = FDC_RECNOTFOUND;
           }

        ctrl_status |= FDC_BUSY | FDC_DRQ;
        ctrl_drq = 1;
        break;

//------------------------------------------------------------------------------
// Read track.
//------------------------------------------------------------------------------
     case FDC_READTRACK:
        if (modio.fdc)
           {
            printf("fdc_cmd: readtrack\n");
            if (modio.level)
               fprintf(modio.log, "fdc_cmd: readtrack\n");
           }     
        break;

//------------------------------------------------------------------------------
// Write track - format a disk track.
//
// based on 256KB/S and 5 disk rotations per second the number of bytes that
// can be placed on a track is:
// 
// Double Density (DD) bytes is 6250 = 250000 / 5 / 8
// Single Density (SD) bytes is 3125 = 250000 / 5 / 8 / 2
//------------------------------------------------------------------------------
     case FDC_WRITETRACK:
        if (modio.fdc)
           {
            printf("fdc_cmd: writetrack\n");
            if (modio.level)
               fprintf(modio.log, "fdc_cmd: writetrack\n");
           }

        buf_index = 0;
        if (ctrl_ddense)
           bytes_left = 6250;
        else   
           bytes_left = 3125;

        write_track_action = 0;
        sector_count = 0;
        data_same = 1;
        data_last = -1;
        data_columns = 0;

        ctrl_status = FDC_BUSY | FDC_DRQ;
        if (disk_iswrprot(&fdc_drive[ctrl_drive].disk))
           {
            ctrl_status |= FDC_WRPROT;
            break;
           }
        ctrl_drq = 1;
        break;

//------------------------------------------------------------------------------
// Forced interrupt.
//
// Used to terminate a multi sector read/write command and to ensure Type I
// status in the status register.
//------------------------------------------------------------------------------
     case FDC_INTERRUPT:
        if (modio.fdc)
           {
            printf("fdc_cmd: interrupt\n");
            if (modio.level)
               fprintf(modio.log, "fdc_cmd: interrupt\n");
           }

        ctrl_status = 0;
        
        if (disk_iswrprot(&fdc_drive[ctrl_drive].disk))
           {
            ctrl_status |= FDC_WRPROT;
            break;
           }

//        if (fdc_drive[ctrl_drive].track == 0)
//           ctrl_status |= FDC_TRACK0;

        if (fdc_drive[ctrl_drive].disk.itype == 0)
           ctrl_status |= FDC_INDEXPULSE;

        bytes_left = 0;
        ctrl_drq = 0;
        ctrl_intrq = 1;
        cmdx = -1;
        break;
    }
}

//==============================================================================
// Read FDC status
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			ctrl_status
//==============================================================================
UINT16 fdc_status (UINT16 port, struct z80PortRead *port_s)
{
 if (modelx.rom)
    return 0;
 
 ctrl_intrq = 0;     // read status clears interrupt

 if (bytes_left != 0)
    {
     buf_index++;
     ctrl_status |= FDC_LOSTDATA;

     if (--bytes_left == 0)
        {
         ctrl_status = FDC_LOSTDATA;
         ctrl_drq = 0;
         ctrl_intrq = 1;
        }
    }

 if (modio.fdc)
    {
     printf("fdc_status: port=0x%02x ctrl_status=0x%02X\n", port, ctrl_status);
     if (modio.level)     
        fprintf(modio.log, "fdc_status: port=0x%02x ctrl_status=0x%02X\n", port, ctrl_status);
    }    
    
 return ctrl_status;
}

//==============================================================================
// Write 1 data byte.
//
// Handles Write sector and Write track FDC commands.
//
// Write sector:
//  The sector(s) is written after a complete sector has been filled.
//
// Write track:
//  Gathers sector header information,  when the count of bytes written is
//  greater than the capacity of one track (SS or DD) a format track command
//  function formats the track using ctrl_side, ctrl_rtrack, and ctrl_ddense
//  values.
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void fdc_data_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 int i;
 
 if (modelx.rom)
    return;

 if (modio.fdc)
    {
     printf("fdc_data_w: port=0x%02x ctrl_rdata=0x%02x\n", port, data);
     if (modio.level)
        fprintf(modio.log, "fdc_data_w: port=0x%02x ctrl_rdata=0x%02x\n", port, data);
    }    

 ctrl_rdata = data;

//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
 if ((cmdx != FDC_WRITESECT) && (cmdx != FDC_WRITETRACK))
    {
     if (disk_iswrprot(&fdc_drive[ctrl_drive].disk))
        ctrl_status |= FDC_WRPROT;
    }    

//------------------------------------------------------------------------------
// Write sector
//------------------------------------------------------------------------------
 if (cmdx == FDC_WRITESECT)
    {
     if (bytes_left != 0)
        {
         if (bytes_sect == 0)
            {
             bytes_sect = fdc_drive[ctrl_drive].disk.secsize;
             buf_index = 0;
            } 

         buf[buf_index++] = data;

         if (--bytes_sect == 0)            	// if end of sector write
            {
             fdc_error = disk_write(&fdc_drive[ctrl_drive].disk, buf, ctrl_side, fdc_drive[ctrl_drive].track, ctrl_rsect, 0);

             bytes_left -= fdc_drive[ctrl_drive].disk.secsize;
         
             if ((bytes_left <= 0) || (fdc_error))
                {
                 cmdx = -1;
                 ctrl_drq = 0;
                 ctrl_intrq = 1;

                 if (fdc_error)
                    {
                     bytes_left = 0;
                     bytes_sect = 0;                 
#ifdef USE_LIBDSK
                     if (fdc_drive[ctrl_drive].disk.itype == DISK_LIBDSK)
                        {
                         switch (fdc_error)
                            {
                             case  DSK_ERR_RDONLY : ctrl_status = FDC_WRPROT;
                                                    break;
                             case  DSK_ERR_NOTRDY : ctrl_status = FDC_NOTREADY;
                                                    break;
                             case  DSK_ERR_NODATA : ctrl_status = FDC_RECNOTFOUND;
                                                    break;
                             case DSK_ERR_DATAERR : ctrl_status = FDC_CRCERROR;
                                                    break;
                             case  DSK_ERR_NOADDR : ctrl_status = FDC_RECNOTFOUND;
                                                    break;                                       
                                          default : ctrl_status = FDC_LOSTDATA;
                                                    break;
                            }
                        }
                     else    
#endif
                        ctrl_status = FDC_RECNOTFOUND;
                    }
                 else   
                    ctrl_status = 0;
                } 
            }
        }
    }
    
//------------------------------------------------------------------------------
// Write track
//
// LibDsk:
// The side used in the sector header is determined by the side1as0 member.
// This value is set directly from the command line or when some disk
// formats are specified.
//------------------------------------------------------------------------------
 if (cmdx == FDC_WRITETRACK)
    {
     switch (write_track_action)
        {
         case 0 : switch (data)
                     {
                      case 0xfe : write_track_action = 1; // receive sector header
                                  sector_header_pos = 0;
                                  sector_count++;
                                  break;
                      case 0xf7 : bytes_left--; // extra count for 2 CRCs
                                  break;
                     }
                  break;
         case 1 : if (sector_header_pos == 1)
                     {
#if USE_LIBDSK
                      if (fdc_drive[ctrl_drive].disk.side1as0)
                         buf[buf_index++] = data;	// use native method
                      else   
                         buf[buf_index++] = ctrl_side;  // force to current physical side
#else
                      buf[buf_index++] = data;
#endif                         
                     }
                  else   
                     buf[buf_index++] = data;
                  if (++sector_header_pos == 4)
                     write_track_action = 0;
                  break;
        }

// show the track write data
     if (modio.fdc_wtd)
        {
         if (data_last == -1)
            {
             data_last = data;
             printf("\nfdc_data_w: Write track data: Drive=%c: DD=%d Track=%d Side=%d\n",
             ctrl_drive + 'A', ctrl_ddense, ctrl_rtrack, ctrl_side);
             if (modio.level)
                fprintf(modio.log, "\nfdc_data_w: Write track data: Drive=%c: DD=%d Track=%d Side=%d\n",
                ctrl_drive + 'A', ctrl_ddense, ctrl_rtrack, ctrl_side);
            } 

         if (data_last == data)
            data_same++;

         if ((data_last != data) || (bytes_left <= 0))
            {
             printf("0x%02xx%-5d", data_last, data_same);
             if (modio.level)
                fprintf(modio.log, "0x%02xx%-5d", data_last, data_same);
             if (((++data_columns % 8) == 0) || (bytes_left <= 0))
                {
                 printf("\n");
                 if (modio.level)
                    fprintf(modio.log, "\n");
                }    
             data_last = data;
             data_same = 1;
            }
        }    

     if (--bytes_left < 0)  // if a full write track completed
        {
// show the sector header information
         if (modio.fdc_wth)
            {
             printf("\nfdc_data_w: Write track header: Drive=%c: DD=%d Track=%d Side=%d\n", ctrl_drive + 'A', ctrl_ddense, ctrl_rtrack, ctrl_side);
             printf("Track  Head  Sect  Size\n");
             for (i = 0; i < sector_count; i++)
                printf("%5d%6d%6d%6d\n", buf[i*4+0], buf[i*4+1], buf[i*4+2], buf[i*4+3] << 8);
             
             if (modio.level)
                 {
                  fprintf(modio.log, "\nfdc_data_w: Write track header: Drive=%c: DD=%d Track=%d Side=%d\n", ctrl_drive + 'A', ctrl_ddense, ctrl_rtrack, ctrl_side);
                  fprintf(modio.log, "Track  Head  Sect  Size\n");
                  for (i = 0; i < sector_count; i++)
                      fprintf(modio.log, "%5d%6d%6d%6d\n", buf[i*4+0], buf[i*4+1], buf[i*4+2], buf[i*4+3] << 8);
                 } 
            }    

         cmdx = -1;
         fdc_error = disk_format_track(&fdc_drive[ctrl_drive].disk, buf,
         ctrl_ddense, ctrl_rtrack, ctrl_side, sector_count);
         if (fdc_error)
            {
#ifdef USE_LIBDSK
             if (fdc_drive[ctrl_drive].disk.itype == DISK_LIBDSK)
                {
                 switch (fdc_error)
                    {
                     case  DSK_ERR_RDONLY : ctrl_status = FDC_WRPROT;
                                            break;
                     case  DSK_ERR_NOTRDY : ctrl_status = FDC_NOTREADY;
                                            break;
                                  default : ctrl_status = FDC_LOSTDATA;
                                            break;
                    }
                }
             else    
#endif
                ctrl_status = FDC_LOSTDATA;
            }
         else   
           ctrl_status = 0;
        }
    }
}

//==============================================================================
// Read 1 data byte from sector
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: UINT16 			ctrl_rdata
//==============================================================================
UINT16 fdc_data_r (UINT16 port, struct z80PortRead *port_s)
{
 if (modelx.rom)
    return 0;

 if (bytes_left != 0)
    {
     ctrl_rdata = buf[buf_index++];
     if (--bytes_left == 0)		// if end of sector read
        {
         ctrl_drq = 0;
         ctrl_intrq = 1;
         ctrl_status &= 0xFF ^ (FDC_BUSY | FDC_DRQ );         
        }
    }

 if (modio.fdc)
    {
     printf("fdc_data_r: port=0x%02x ctrl_rdata=0x%02x\n", port, ctrl_rdata & 0xFF);
     if (modio.level)
        fprintf(modio.log, "fdc_data_r: port=0x%02x ctrl_rdata=0x%02x\n", port, ctrl_rdata & 0xFF);
    }    

 return ctrl_rdata & 0xFF;
}

//==============================================================================
// Set track number
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void fdc_track_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.rom)
    return;

 if (modio.fdc)
    {
     printf("fdc_track_w: port=0x%02x ctrl_rtrack=0x%02x\n", port, data);
     if (modio.level)
        fprintf(modio.log, "fdc_track_w: port=0x%02x ctrl_rtrack=0x%02x\n", port, data);
    }    
 ctrl_rtrack = data;
}

//==============================================================================
// Get track number
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: UINT16			ctrl_rtrack
//==============================================================================
UINT16 fdc_track_r (UINT16 port, struct z80PortRead *port_s)
{
 if (modelx.rom)
    return 0;

 if (modio.fdc)
    {
     printf("fdc_track_r: port=0x%02x ctrl_rtrack=0x%02x\n", port, ctrl_rtrack);
     if (modio.level)
        fprintf(modio.log, "fdc_track_r: port=0x%02x ctrl_rtrack=0x%02x\n", port, ctrl_rtrack);
    }    
 return ctrl_rtrack;
}

//==============================================================================
// Set sector number
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void fdc_sect_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.rom)
    return;

 if (modio.fdc)
    {
     printf("fdc_sect_w: port=0x%02x ctrl_rsect=0x%02x\n", port, data);
     if (modio.level)
        fprintf(modio.log, "fdc_sect_w: port=0x%02x ctrl_rsect=0x%02x\n", port, data);
    }    

 ctrl_rsect = data;
}

//==============================================================================
// Get sector number
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: UINT16			ctrl_rsect
//==============================================================================
UINT16 fdc_sect_r (UINT16 port, struct z80PortRead *port_s)
{
 if (modelx.rom)
    return 0;

 if (modio.fdc)
    {
     printf("fdc_sect_r: port=0x%02x ctrl_rsect=0x%02x\n", port, ctrl_rsect);
     if (modio.level)
        fprintf(modio.log, "fdc_sect_r: port=0x%02x ctrl_rsect=0x%02x\n", port, ctrl_rsect);
    }    

 return ctrl_rsect;
}

//==============================================================================
// Set drive, side and density
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void fdc_ext_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.rom)
    return;

 if (modio.fdc)
    {
     printf("fdc_ext_w: port=0x%02x ctrl_drive/side/ddense=0x%02x\n", port, data);
     if (modio.level)
        fprintf(modio.log, "fdc_ext_w: port=0x%02x ctrl_drive/side/ddense=0x%02x\n", port, data);
    }    

 ctrl_drive = data & 0x03;
 ctrl_side = (data & 0x04) >> 2;
 ctrl_ddense = (data & 0x08) >> 3;

 if (fdc_drive[ctrl_drive].disk.itype == 0)
    ctrl_status |= FDC_INDEXPULSE;              // set index pulse if no disk

 gui_status_update(ctrl_drive + 'A');
}

//==============================================================================
// Get intrq and drq
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			0x80 if intrq or drq set, else 0
//==============================================================================
UINT16 fdc_ext_r (UINT16 port, struct z80PortRead *port_s)
{
 int status;

 if (modelx.rom)
    return 0;

 if (ctrl_intrq | ctrl_drq) 
    status = 0x80;
 else
    status = 0x00;

 if (modio.fdc)
    {
     printf("fdc_ext_r: port=0x%02x (ctrl_intrq | ctrl_drq) status=0x%02x\n", port, status);
     if (modio.level)
        fprintf(modio.log, "fdc_ext_r: port=0x%02x (ctrl_intrq | ctrl_drq) status=0x%02x\n", port, status);
    }    

// force returning a status value of 0,  used to simulate no disk in drive
 if (fdc_status0)
    {
     fdc_status0--;
     return 0;
    }

 return status;
}
